class Channel(int):
    """
    This class is introduced to distinguish numerical algorithm parameter from choose chanel.
    In autogenerated interface field with this type limits input values to number of current image channels
    """

    def __str__(self):
        return str(self + 1)
